// =========================================
// The special Minisat constraint that makes
// Sure that the candidate automaton does not
// accept a language too small.
// =========================================
#include "AcceptEnoughConstraint.hpp"
#include "ReferenceAutomatonBitMaskTypeTrie.hpp"
#include <boost/tuple/tuple.hpp>
#include <list>
#include <set>

//===================================
// Minisat functions
//===================================
void Solver::addAcceptEnoughConstraint(const NBW &neg, int targetNofStates, Var ***transitionVars, Var *acceptanceVars, uint bound) {
    if (ok) {
        AcceptEnoughConstraint* c; ok = AcceptEnoughConstraint_new(*this, neg, targetNofStates, transitionVars, acceptanceVars, bound, c);
        if (c != NULL) constrs.push(c);
    }
}

bool AcceptEnoughConstraint_new(Solver& S, const NBW &neg, int targetNofStates, Var ***transitionVars, Var *acceptanceVars, uint bound, AcceptEnoughConstraint*& out_constr) {
    out_constr = new AcceptEnoughConstraint(S,neg,targetNofStates,transitionVars,acceptanceVars, bound);
    return true;
}

//===================================
// Constructor & Destruction helper
//===================================
AcceptEnoughConstraint::AcceptEnoughConstraint(Solver& _S, const NBW &_pos, int _targetNofStates, Var ***_transitionVars, Var *_acceptanceVars, uint _bound) : solver(_S), pos(_pos), targetNofStates(_targetNofStates), transitionVars(_transitionVars), acceptanceVars(_acceptanceVars), chosenBound(_bound), memoryRecycler(targetNofStates)
{
    // Register into watch list
    for (uint i=0;i<targetNofStates;i++) {
        for (uint j=0;j<pos.getSizeSigma();j++) {
            for (uint k=0;k<targetNofStates;k++) {
                solver.watches[index(Lit(transitionVars[i][j][k],true))].push(this);
            }
        }
        solver.watches[index(Lit(acceptanceVars[i],true))].push(this);
    }
}

void AcceptEnoughConstraint::remove(Solver& UNUSED(S), bool just_dealloc) {
    if (!just_dealloc) {
        for (uint i=0;i<targetNofStates;i++) {
            for (uint j=0;j<pos.getSizeSigma();j++) {
                for (uint k=0;k<targetNofStates;k++) {
                    removeWatch(solver.watches[index(Lit(transitionVars[i][j][k],true))],this);
                }
            }
            removeWatch(solver.watches[index(Lit(acceptanceVars[i],true))],this);
        }
    }
    delete this;
}

//===================================
// Is called whenever a new decision has been made
//===================================
bool AcceptEnoughConstraint::propagate(Solver& UNUSED(S), Lit UNUSED(p), bool& keep_watch) {
    keep_watch = true;
    memoryRecycler.clear();

    return performCheck(chosenBound);
}

bool AcceptEnoughConstraint::performCheck(uint bound) {

    // Compute successors
    ReferenceAutomatonBitMaskTypeTrie alreadyVisited(targetNofStates,bound+1); // We always include the bound as a possible number

    // Make initial node
    int *init = memoryRecycler.getNew();
    init[0] = (pos.isAccepting(0) && (solver.assigns[acceptanceVars[0]]==l_False_char))?bound-1:bound;
    for (uint i=1;i<targetNofStates;i++) {
        init[i] = 0;
    }
    #ifdef TRACK_CAUSE_IN_NOT_TOO_MUCH_CONSTRAINT
    todoList.push_back(boost::make_tuple(init,0,0));
    #else
    todoList.push_back(boost::make_tuple(init,0));
    #endif

    while (todoList.size()>0) {
        #ifdef TRACK_CAUSE_IN_NOT_TOO_MUCH_CONSTRAINT
        boost::tuple<int*,int,NotTooMuchCauseTrackingType > current = todoList.back();
        #else
        boost::tuple<int*,int > current = todoList.back();
        #endif

        todoList.pop_back();

        // Positioning...
        /*
        std::cout << "Visit: Candidate: {";
        for (uint i=0;i<targetNofStates;i++) {
            if (i!=0) std::cout << ",";
            std::cout << current.get<0>()[i];
        }
        std::cout << "}, Ref: " << current.get<1>() << "\n";*/

        // Evaluate successors...
        MaskReference oldPos = alreadyVisited.find(current.get<0>());
        bool hasSet = oldPos.isSet(current.get<1>());
        if (!hasSet) {

            oldPos.insert(current.get<1>());

             // Check if we found a conflict
            {
                uint i=0;
                while ((i<targetNofStates) && (current.get<0>()[i]==0)) i++;
                if (i==targetNofStates) {
                    // Found conflict
                    todoList.clear();
                    #ifdef TRACK_CAUSE_IN_NOT_TOO_MUCH_CONSTRAINT
                    lastCause = current.get<2>();
                    #endif
                    return false;
                 }
            }

            // Now compute successors for every symbol
            for (int k=pos.getSizeSigma()-1;k>=0;k--) {
                int* next = memoryRecycler.getNew();
                for (uint i=0;i<targetNofStates;i++) next[i] = 0;

                // Find possible transitions
                for (int i=targetNofStates-1;i>=0;i--) {
                    int v = current.get<0>()[i];
                    if (v!=0) {
                        Var* tVPos = &(transitionVars[i][k][0]);
                        for (uint j=0;j<targetNofStates;j++) {
                            char trans = solver.assigns[*(tVPos++)];
                            if (trans!=l_False_char) {
                                next[j] = std::max(next[j],v);
                            }
                       }
                    }
                    assert(next[i] >= 0);
                }

                // Find possible visits to accepting states and prepare successor
                // vectors for the two possibilities of a next state being accepting or not
                int* nextAcc = memoryRecycler.getNew();
                for (int i=targetNofStates-1;i>=0;i--) {
                    char acc = solver.assigns[acceptanceVars[i]];
                    if ((acc!=l_False_char) && (next[i]!=0)) {
                        next[i] = bound;
                        nextAcc[i] = bound;
                    } else {
                        nextAcc[i] = std::max(0,next[i]-1);
                    }
                }

                // Insert new stuff into todoList
                for (uint i=0;i<pos.getNofStates();i++) {
                    if (pos.hasTransition(current.get<1>(),k,i)) {
                        #ifdef TRACK_CAUSE_IN_NOT_TOO_MUCH_CONSTRAINT
                        if (pos.isAccepting(i)) {
                            todoList.push_back(boost::make_tuple(nextAcc,i,current.get<2>() | ((NotTooMuchCauseTrackingType)1<<(NotTooMuchCauseTrackingType)k)));
                        } else {
                            todoList.push_back(boost::make_tuple(next,i,current.get<2>() | ((NotTooMuchCauseTrackingType)1<<(NotTooMuchCauseTrackingType)k)));
                        }
                        #else
                        if (pos.isAccepting(i)) {
                            todoList.push_back(boost::make_tuple(nextAcc,i));
                        } else {
                            todoList.push_back(boost::make_tuple(next,i));
                        }
                        #endif
                    }
                }
            }
        }
    }

    return true;
}

//===================================
// Compute a reason
//===================================
void AcceptEnoughConstraint::calcReason(Solver& UNUSED(S), Lit p, vec<Lit>& out_reason) {
    // Reasons here are always simple: All variables set to "0" are bad.
    if (p==lit_Error) {
        std::cerr << "AcceptNotTooMuchConstraint::calcReason(Solver& S, Lit p, vec<Lit>& out_reason) - Got lit_undef";
        throw 42;
    } else if (p==lit_Undef) {
        vec<char> &assigns = solver.assigns;
        //uint nof = 0;
        // std::cout << lastCause << "\n";
        for (uint i=0;i<targetNofStates;i++) {
            for (uint k=0;k<pos.getSizeSigma();k++) {
                #ifdef TRACK_CAUSE_IN_NOT_TOO_MUCH_CONSTRAINT
                if ((lastCause & ((NotTooMuchCauseTrackingType)1<<(NotTooMuchCauseTrackingType)k))!=0) {
                #endif
                for (uint j=0;j<targetNofStates;j++) {
                    Var current = transitionVars[i][k][j];
                    if (assigns[current] == l_False_char) {
                        out_reason.push(Lit(current,true));
                        // nof++;
                    }
                }
                #ifdef TRACK_CAUSE_IN_NOT_TOO_MUCH_CONSTRAINT
                }
                #endif
            }
            Var current = acceptanceVars[i];
            if (assigns[current] == l_False_char) {
                out_reason.push(Lit(current,true));
                //nof++;
            }
        }

        /*for (uint i=0;i<assigns.size();i++) {
            Var current = i;
            if (assigns[current] == l_False_char) {
                nof++;
                out_reason.push(Lit(current,true));
            }
        }*/
        //std::cout << nof << std::endl;
    } else {
        std::cerr << "AcceptNotTooMuchConstraint::calcReason(Solver& S, Lit p, vec<Lit>& out_reason) - Got non-conflict literal";
        throw 42;
    }

}
