// =========================================
// The special Minisat clause that makes
// Sure that the candidate automaton does not
// accept a language too large.
// =========================================
#include "AcceptNotTooMuchConstraint.hpp"

#include "Minisat/Solver.h"
#include "NBW.hpp"
#include <tr1/unordered_map>

//===================================
// Minisat functions
//===================================
void Solver::addAcceptNotTooMuchConstraint(const NBW &neg, int targetNofStates, Var ***transitionVars, Var *acceptanceVars) {
    if (ok) {
        AcceptNotTooMuchConstraint* c; ok = AcceptNotTooMuchConstraint_new(*this, neg, targetNofStates, transitionVars, acceptanceVars, c);
        if (c != NULL) constrs.push(c);
    }
}

bool AcceptNotTooMuchConstraint_new(Solver& S, const NBW &neg, int targetNofStates, Var ***transitionVars, Var *acceptanceVars, AcceptNotTooMuchConstraint*& out_constr) {
    out_constr = new AcceptNotTooMuchConstraint(S,neg,targetNofStates,transitionVars,acceptanceVars);
    return true;
}


//===================================
// Constructor & destructor
//===================================
AcceptNotTooMuchConstraint::AcceptNotTooMuchConstraint(Solver &_solver, const NBW &_neg, uint _targetNofStates, Var ***_transitionVars, Var *_acceptanceVars) : solver(_solver), neg(_neg), targetNofStates(_targetNofStates), transitionVars(_transitionVars), acceptanceVars(_acceptanceVars)
{
    // Register into watch list
    for (uint i=0;i<targetNofStates;i++) {
        for (uint j=0;j<neg.getSizeSigma();j++) {
            for (uint k=0;k<targetNofStates;k++) {
                solver.watches[index(Lit(transitionVars[i][j][k],false))].push(this);
            }
        }
        solver.watches[index(Lit(acceptanceVars[i],false))].push(this);
    }

    // Generate reason table
    dfsVisitedMode = new int[targetNofStates*3*neg.getNofStates()];
    reasons1 = new Lit[targetNofStates*3*neg.getNofStates()];
    reasons2 = new Lit[targetNofStates*3*neg.getNofStates()];
    prevReasonPointers = new int[targetNofStates*3*neg.getNofStates()];

}

AcceptNotTooMuchConstraint::~AcceptNotTooMuchConstraint() {
    delete[] dfsVisitedMode;
    delete[] reasons1;
    delete[] reasons2;
    delete[] prevReasonPointers;
}

void AcceptNotTooMuchConstraint::remove(Solver& UNUSED(S), bool just_dealloc) {
    if (!just_dealloc) {
        for (uint i=0;i<targetNofStates;i++) {
            for (uint j=0;j<neg.getSizeSigma();j++) {
                for (uint k=0;k<targetNofStates;k++) {
                    removeWatch(solver.watches[index(Lit(transitionVars[i][j][k],false))],this);
                }
            }
            removeWatch(solver.watches[index(Lit(acceptanceVars[i],false))],this);
        }
    }
    delete this;
}

//===================================
// Is called whenever a new decision has been made
//===================================
bool AcceptNotTooMuchConstraint::propagate(Solver& UNUSED(S), Lit UNUSED(p), bool& keep_watch) {
    keep_watch=true;
    /*std::cout << "Set: ";
    for (uint i=0;i<targetNofStates;i++) {
        for (uint j=0;j<neg.getSizeSigma();j++) {
            for (uint k=0;k<targetNofStates;k++) {
                 char valueSet = solver.assigns[transitionVars[i][j][k]];
                 if (valueSet == toInt(l_True)) {
                     std::cout << i << "-(" << j << ")->" << k << ", ";
                 }
            }
        }
        char valueSet = solver.assigns[acceptanceVars[i]];
        if (valueSet == toInt(l_True)) {
                 std::cout << "acc(" << i << "), ";
        }
     }
    std::cout << std::endl;*/

    // Use the double-DFS algorithm from [CVWY91].
    // See http://www7.in.tum.de/~esparza/Talks/GAMES.pdf for details

    // Blank the buffer
    memset(dfsVisitedMode,0,sizeof(int)*neg.getNofStates()*3*targetNofStates);
    acceptingStatesInPostOrder.clear();
    reasons1[posInTable(0,0,0)]=lit_Undef;
    reasons2[posInTable(0,0,0)]=lit_Undef;
    prevReasonPointers[posInTable(0,0,0)]=-1;

    // First round
    dfsFirstRound(posInTable(0,0,0));
    lastReason.clear();

    // Second round - Search for a look
    for (uint i=0;i<acceptingStatesInPostOrder.size();i++) {
        uint current = acceptingStatesInPostOrder[i];
        if (dfsSecondRound(current, current)) return false;
    }

    // Everything is fine.
    return true;
};


//===================================
// Double-DFS: First iteration
//===================================
void AcceptNotTooMuchConstraint::dfsFirstRound(uint currentNode) {
    
    // Add to the visitor list
    dfsVisitedMode[currentNode]=1;
    uint candidate = (currentNode / 3) % targetNofStates;
    uint reference = currentNode / 3 / targetNofStates;
    uint thirdComponent = currentNode % 3;

    // std::cout << "First round: " << candidate << ", " << reference << ", " << thirdComponent << std::endl;

    // Find post-States
    for (uint k=0;k<neg.getSizeSigma();k++) {
        for (uint c=0;c<targetNofStates;c++) {
            Var which = transitionVars[candidate][k][c];
            char what = solver.assigns[which];
            if (what==l_True_char) {
                for (uint r=0;r<neg.getNofStates();r++) {
                    if (neg.hasTransition(reference,k,r)) {
                        if (thirdComponent==2) {
                            uint target = posInTable(r,c,0);
                            if (dfsVisitedMode[target]==0) {
                                // Visit!
                                reasons1[target] = Lit(which,false);
                                reasons2[target] = lit_Undef;
                                prevReasonPointers[target] = currentNode;
                                dfsFirstRound(target);
                            }
                        } else if (thirdComponent==0) {
                            char accSet = solver.assigns[acceptanceVars[c]];
                            uint targetThirdComponent=(accSet==l_True_char)?1:0;
                            uint target = posInTable(r,c,targetThirdComponent);
                            if (dfsVisitedMode[target]==0) {
                                reasons2[target] = (accSet==l_True_char)?Lit(acceptanceVars[c],false):lit_Undef;
                                reasons1[target] = Lit(which,false);
                                prevReasonPointers[target] = currentNode;
                                dfsFirstRound(target);
                            }
                        } else {
                            uint targetThirdComponent=neg.isAccepting(r)?2:1;
                            uint target = posInTable(r,c,targetThirdComponent);
                            if (dfsVisitedMode[target]==0) {
                                reasons2[target] = lit_Undef;
                                reasons1[target] = Lit(which,false);
                                prevReasonPointers[target] = currentNode;
                                dfsFirstRound(target);
                            }
                        }
                    }
                }
            }
        }
    }

    // Done. Add to accepting state list if applicable
    if (thirdComponent==2) {
        acceptingStatesInPostOrder.push_back(currentNode);
    }
}


//===================================
// Double-DFS: Second iteration - Returns true if loop has been found
//===================================
bool AcceptNotTooMuchConstraint::dfsSecondRound(uint basisNode, uint currentNode) {

    // Add to the visitor list
    dfsVisitedMode[currentNode]=2;
    uint candidate = (currentNode / 3) % targetNofStates;
    uint reference = currentNode / 3 / targetNofStates;
    uint thirdComponent = currentNode % 3;

    // Find post-States
    for (uint k=0;k<neg.getSizeSigma();k++) {
        for (uint c=0;c<targetNofStates;c++) {
            Var which = transitionVars[candidate][k][c];
            char what = solver.assigns[which];
            if (what==l_True_char) {
                for (uint r=0;r<neg.getNofStates();r++) {
                    if (neg.hasTransition(reference,k,r)) {
                        uint targetThirdComponent;
                        Lit additionalConflict = lit_Undef;
                        if (thirdComponent==2) {
                            targetThirdComponent=0;
                        } else if (thirdComponent==0) {
                            char accSet = solver.assigns[acceptanceVars[c]];
                            targetThirdComponent=(accSet==l_True_char)?1:0;
                            if (accSet==l_True_char) {
                                additionalConflict = Lit(acceptanceVars[c],false);
                            }
                        } else {
                            targetThirdComponent=neg.isAccepting(r)?2:1;
                        }
                        uint nextNode = posInTable(r,c,targetThirdComponent);
                        if (nextNode==basisNode) {
                            // Copy the basis reason
                            if (additionalConflict!=lit_Undef) lastReason.insert(additionalConflict);
                            lastReason.insert(Lit(which,false));
                            int pointer=basisNode;
                            while (pointer!=-1) {
                                Lit a = reasons1[pointer];
                                if (a!=lit_Undef) lastReason.insert(a);
                                Lit b = reasons2[pointer];
                                if (b!=lit_Undef) lastReason.insert(b);
                                pointer = prevReasonPointers[pointer];
                            }
                            // Return conflict
                            return true;
                        }

                        if (dfsVisitedMode[nextNode]!=2) {
                            if (dfsSecondRound(basisNode,nextNode)) {
                                if (additionalConflict!=lit_Undef) lastReason.insert(additionalConflict);
                                lastReason.insert(Lit(which,false));
                                return true;
                            }
                        }
                    }
                }
            }
        }
    }
    return false;
}


//===================================
// Get the reason for a conflict
//===================================
void AcceptNotTooMuchConstraint::calcReason(Solver& UNUSED(S), Lit p, vec<Lit>& out_reason) {
    if (p==lit_Error) {
        std::cerr << "AcceptNotTooMuchConstraint::calcReason(Solver& S, Lit p, vec<Lit>& out_reason) - Got lit_undef";
        throw 42;
    } else if (p==lit_Undef) {
        for (std::set<Lit>::iterator it = lastReason.begin();it!=lastReason.end();it++) {
            out_reason.push(*it);
#ifndef NDEBUG
            char valueSet = solver.assigns[var(*it)];
            if (valueSet != l_True_char) {
                std::cerr << "Error: NotTooMuchConstraint: Learning incorrect stuff!" << std::endl;
                throw 666;
            }
#endif
        }
    } else {
        std::cerr << "AcceptNotTooMuchConstraint::calcReason(Solver& S, Lit p, vec<Lit>& out_reason) - Got non-conflict literal";
        throw 42;
    }
}
