//========================================================
// A class storing a non-determinstic Büchi word automaton
//========================================================

#ifndef NBW_H
#define NBW_H

#include <string>
#include <iostream>
#include <sstream>
#include <set>
#include <vector>
#include <string>
#include <stdlib.h>
#include <boost/algorithm/string/regex.hpp>
#include <boost/algorithm/string/finder.hpp>
#include <boost/algorithm/string/trim.hpp>
#include <boost/utility.hpp>
#include <boost/tuple/tuple.hpp>
#include <boost/tuple/tuple_comparison.hpp>

//==============================================================
// Expect functions for parsing command line output
//==============================================================
class GeneralException {
        std::string message;
public:
        GeneralException(const std::string &msg) : message(msg) {};
        virtual ~GeneralException() {};
        const std::string getMessage() const { return message; };
};

inline void expect(std::istream &in, const std::string &what) {
        char *c = new char[what.length()+1];
        in.get(c,what.length()+1);
        std::cout.flush();
        if (c!=what) {
                std::ostringstream msg;
                msg << "Expectation Failure, expected: '" << what << "' but found '" <<c<<"'.";
                delete[] c;
                throw GeneralException(msg.str());
        }
        delete[] c;
}

//==============================================================
// The NBW class
//==============================================================
class PreNBW;

class NBW : private boost::noncopyable {
private:
    uint nofStates;
    uint sizeSigma;
    bool *accepting;
    bool ***transitions;
    std::vector<std::string> alphabet;

public:
    NBW(const PreNBW &source);
    NBW(uint _nofStates, uint _sizeSigma);
    ~NBW();
    void print();
    inline uint getSizeSigma() const { return sizeSigma; }
    inline uint getNofStates() const { return nofStates; }
    inline bool hasTransition(uint from, uint symbol, uint to) const { return transitions[from][symbol][to]; }
    inline bool isAccepting(uint state) const { return accepting[state]; }
    inline void setAccepting(uint stateNr, bool acc) { accepting[stateNr] = acc; }
    inline void setTransition(uint i, uint j, uint k, bool t) { transitions[i][j][k] = t; }
    void copyAlphabetFrom(const NBW &other) { alphabet = other.alphabet; }
};

//==============================================================
// The PreNBW class
// This differs from the NBW class by the fact that the alphabet has
// already been "flattened" for NBW
//==============================================================
class PreNBW {
private:
    std::set<std::string> states;
    std::set<std::string> acceptingstates;
    std::set<boost::tuple<std::string,std::vector<std::string>,std::string> > transitions;
    std::set<std::string> alphabet;
    std::string initialState;
public:
    PreNBW(const char *inputNeverClaimFilename);
    void addSPINNotationTransition(std::string from, std::string to, std::string what);
    void addToAlphabet(std::set<std::string> stuff);
    std::set<std::string> getAlphabet() { return alphabet; }
    friend class NBW;
};

#endif // NBW_H
