// =========================================
// The special Minisat constraint that breaks
// symmetry partially
// =========================================

#include "SymmetryBreaker.hpp"

//===================================
// Minisat functions
//===================================
void Solver::addLexicographicConstraint(Var *_equality1, Var *_equality2, Lit _smaller1, Lit smaller2, int size) {
    if (ok) {
        LexicographicConstraint* c; ok = LexicographicConstraint_new(*this, _equality1, _equality2,  _smaller1,  smaller2, size, c);
        if (c != NULL) constrs.push(c);
    }
}

bool LexicographicConstraint_new(Solver& S, Var *_equality1, Var *_equality2, Lit _smaller1, Lit smaller2, int size, LexicographicConstraint*& out_constr) {
    out_constr = new LexicographicConstraint(S, _equality1, _equality2,  _smaller1,  smaller2, size);
    return true;
}

//===================================
// Constructors and destructors
//===================================
void make_symmetry_breaking_constraints(Solver& solver, uint targetNofStates, uint sizeSigma, Var ***transitionVars, Var *acceptanceVars, uint whichState)  {

    // Allocate fast lookup tables;
    uint nofVarsInRow = (targetNofStates*2-1)*sizeSigma+1;
    Var leftRow[nofVarsInRow];
    Var rightRow[nofVarsInRow];

    // Fill the forward lookup tables
    uint currentPosInRow = 0;
    leftRow[currentPosInRow] = acceptanceVars[whichState];
    rightRow[currentPosInRow++] = acceptanceVars[whichState+1];
    for (uint i=0;i<targetNofStates;i++) {
        if (i!=whichState) {
            for (uint k=0;k<sizeSigma;k++) {
                 leftRow[currentPosInRow] = transitionVars[whichState][k][i];
                 rightRow[currentPosInRow++] = transitionVars[whichState+1][k][(i==whichState+1)?whichState:i];
                 leftRow[currentPosInRow] = transitionVars[i][k][whichState];
                 rightRow[currentPosInRow++] = transitionVars[(i==whichState+1)?whichState:i][k][whichState+1];
            }
        } else {
            for (uint k=0;k<sizeSigma;k++) {
                leftRow[currentPosInRow] = transitionVars[whichState][k][whichState];
                rightRow[currentPosInRow++] = transitionVars[whichState+1][k][whichState+1];
            }
        }
    }

    assert(nofVarsInRow==currentPosInRow);

    // Build the lexicographic stuff...
    for (uint i=0;i<nofVarsInRow;i++) {
        if (i==0) {
            vec<Lit> v;
            v.push(Lit(leftRow[i],true));
            v.push(Lit(rightRow[i],false));
            solver.addClause(v);
        } else {
            Var *first = new Var[i];
            Var *second = new Var[i];
            for (uint j=0;j<i;j++) {
                first[j] = leftRow[j];
                second[j] = rightRow[j];
            }
            solver.addLexicographicConstraint(first,second, Lit(leftRow[i],true), Lit(rightRow[i],false), i);
        }

    }
}


//==============================
// Constsructor & destructor
//==============================
LexicographicConstraint::LexicographicConstraint(Solver& solver, Var *_equality1, Var *_equality2, Lit _smaller1, Lit _smaller2, int _size) : equality1(_equality1), equality2(_equality2), smaller1(_smaller1), smaller2(_smaller2), size(_size) {
    assert(size>0);
    equalities = new bool[size];
    nofEqualities=0;
    posMapper = new int[solver.nVars()];
    for (int i=0;i<size;i++) {
        equalities[i] = false;
        solver.watches[index(Lit(equality1[i],false))].push(this);
        solver.watches[index(Lit(equality1[i],true))].push(this);
        solver.watches[index(Lit(equality2[i],false))].push(this);
        solver.watches[index(Lit(equality2[i],true))].push(this);
        posMapper[equality1[i]] = i;
        posMapper[equality2[i]] = i;
    }
    solver.watches[index(~smaller1)].push(this);
    solver.watches[index(~smaller2)].push(this);
}

LexicographicConstraint::~LexicographicConstraint() {
    delete[] equality1;
    delete[] equality2;
    delete[] posMapper;
    delete[] equalities;
}

void LexicographicConstraint::remove    (Solver& solver, bool just_dealloc) {
    if (!just_dealloc) {
        for (int i=0;i<size;i++) {
            removeWatch(solver.watches[index(Lit(equality1[i],false))],this);
            removeWatch(solver.watches[index(Lit(equality1[i],true))],this);
            removeWatch(solver.watches[index(Lit(equality2[i],false))],this);
            removeWatch(solver.watches[index(Lit(equality2[i],true))],this);

        }
        removeWatch(solver.watches[index(~smaller1)],this);
        removeWatch(solver.watches[index(~smaller2)],this);
    }
    delete this;
}

bool LexicographicConstraint::propagate (Solver& S, Lit p, bool& keep_watch) {
    keep_watch = true;

    assert(nofEqualities>=0);
    assert(nofEqualities<=size);

    S.undos[var(p)].push(this);

    if (p==~smaller1) {
        if (nofEqualities==size) {
            // std::cout << "ENQ" << std::endl;
#ifndef NDEBUG
            for (int i=0;i<size;i++) {
              assert(S.value(equality1[i])!=l_Undef);
              assert(S.value(equality2[i])!=l_Undef);
            }
#endif
            return S.enqueue(smaller2, this);
        }
        return true;
    } else if (p==~smaller2) {
        if (nofEqualities==size) {
            // std::cout << "ENQ" << std::endl;
#ifndef NDEBUG
            for (int i=0;i<size;i++) {
              assert(S.value(equality1[i])!=l_Undef);
              assert(S.value(equality2[i])!=l_Undef);
            }
#endif
            return S.enqueue(smaller1, this);
        }
        return true;
    }

    int pos = posMapper[var(p)];

    bool newEqual = false;
    lbool a = S.value(equality1[pos]);
    lbool b = S.value(equality2[pos]);
    if (a!=l_Undef) {
        if (a==b) newEqual = true;
    }

    if (equalities[pos]!=newEqual) {
#ifndef NDEBUG
        // std::cout << "U" << std::endl;
        int b = 0;
        for (int i=0;i<size;i++) {
            if (equalities[i]) b++;
        }
        assert(nofEqualities==b);
#endif

        equalities[pos]=newEqual;
        assert(nofEqualities>=0);
        assert(nofEqualities<=size);
        // std::cout << (newEqual?"N":"O") << size << std::endl;
        if (newEqual) nofEqualities++; else nofEqualities--;
        assert(nofEqualities>=0);
        assert(nofEqualities<=size);
#ifndef NDEBUG
        // std::cout << "U" << std::endl;
        b = 0;
        for (int i=0;i<size;i++) {
            if (equalities[i]) b++;
        }
        assert(nofEqualities==b);
#endif
    }
    return true;
}

void LexicographicConstraint::calcReason(Solver& solver, Lit p, vec<Lit>& out_reason) {
    // Copy all equality stuff
    for (int i=0;i<size;i++) {
        assert(solver.value(equality1[i])!=l_Undef);
        assert(solver.value(equality2[i])!=l_Undef);
        out_reason.push(Lit(equality1[i],(solver.value(equality1[i])==l_False)));
        out_reason.push(Lit(equality2[i],(solver.value(equality2[i])==l_False)));
    }

    // Now see what we have
    if (p==lit_Undef) {
        assert(solver.value(smaller1)==l_False);
        assert(solver.value(smaller2)==l_False);
        out_reason.push(~smaller1);
        out_reason.push(~smaller2);
        return;
    } else if (p==smaller1) {
        assert(solver.value(smaller2)==l_False);
        out_reason.push(~smaller2);
        return;
    }
    assert(p==smaller2);
    assert(solver.value(smaller1)==l_False);
    out_reason.push(~smaller1);
}

void LexicographicConstraint::undo(Solver& S, Lit p) {

    if (p==~smaller1) return;
    if (p==~smaller2) return;

    int pos = posMapper[var(p)];

    bool newEqual = false;
    lbool a = S.value(equality1[pos]);
    lbool b = S.value(equality2[pos]);
    if (a!=l_Undef) {
        if (a==b) newEqual = true;
    }
    if (equalities[pos]!=newEqual) {
#ifndef NDEBUG
        // std::cout << "U" << std::endl;
        int b = 0;
        for (int i=0;i<size;i++) {
            if (equalities[i]) b++;
        }
        assert(nofEqualities==b);
#endif

        equalities[pos]=newEqual;
        // std::cout << (newEqual?"N":"O") << size << std::endl;
        if (!newEqual) nofEqualities--; else nofEqualities++;
        assert(nofEqualities>=0);
        assert(nofEqualities<size);
#ifndef NDEBUG
        // std::cout << "U" << std::endl;
        b = 0;
        for (int i=0;i<size;i++) {
            if (equalities[i]) b++;
        }
        assert(nofEqualities==b);
#endif
    }

}
