// ================================================
// SAT-Based bounded NBW minimizer
//
// (C) 2010 by Ruediger Ehlers
// ================================================

/**
MiniSat -- Copyright (c) 2003-2005, Niklas Een, Niklas Sorensson

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**/

#include <iostream>
#include <sstream>
#include <stdlib.h>
#include "NBW.hpp"
#include "Minisat/Solver.h"
#include "SymmetryBreaker.hpp"
#include "ReferenceAutomatonBitMaskTypeTrie.hpp"
#include "AcceptEnoughConstraint.hpp"

/***
 *  Prints the tool usage.
 ***/
void usage() {
    std::cerr << "Usage: \nNBWSat <input directory> <targetNofStates> <bound>\n";
}

/***
 * Performs SAT-solving based minimization after the NBWs have been read
 */
void minimize(const NBW &pos, const NBW &neg, uint bound, uint targetNofStates) {

    // Allocate solver & Variables
    Solver solver;
    solver.verbosity = 1;
    Var ***transitionVars = new Var**[targetNofStates];
    Var acceptanceVars[targetNofStates];
    for (uint i=0;i<targetNofStates;i++) {
        transitionVars[i] = new Var*[pos.getSizeSigma()];
        for (uint j=0;j<pos.getSizeSigma();j++) {
            transitionVars[i][j] = new Var[targetNofStates];
            for (uint k=0;k<targetNofStates;k++) {
                transitionVars[i][j][k] = solver.newVar();
                //std::cout << "Transition var: " << i << "," << j << "," << k << ": " << transitionVars[i][j][k] << "\n";
            }
        }
    }
    for (uint i=0;i<targetNofStates;i++) {
        acceptanceVars[i] = solver.newVar();
        //std::cout << "Acceptance var: " << i << ": " << acceptanceVars[i] << "\n";
    }

    // Add basic clauses
    solver.addAcceptNotTooMuchConstraint(neg,targetNofStates,transitionVars,acceptanceVars);
    solver.addAcceptEnoughConstraint(pos,targetNofStates,transitionVars,acceptanceVars,bound);
    for (uint i=2;i<targetNofStates;i++) {
        make_symmetry_breaking_constraints(solver,targetNofStates,pos.getSizeSigma(),transitionVars,acceptanceVars,i-1);
    }

    // Do some symmetry breaking
    /*for (uint b=0;b<targetNofStates;b++) {
        for (uint k=0;k<pos.getSizeSigma();k++) {
            for (uint i=b+2;i<targetNofStates;i++) {
                vec<Lit> reference;
                // Already used...
                for (uint b2=0;b2<=b;b2++) {
                    for (uint k2=0;k2<((b2==b)?k+1:pos.getSizeSigma());k2++) {
                        //std::cout << "b2: " << b2 << ", k2: " << k2 << ", i-1: " << i-1 << "\n";
                        reference.push(Lit(transitionVars[b2][k2][i-1],false));
                    }
                }
                // Is in use
                reference.push(Lit(transitionVars[b][k][i],true));
                //std::cout << "b:" << b << ", k: " << k << ", i: " << i << "\n";
                solver.addClause(reference);
                //std::cout << "done" << std::endl;

            }
        }
    }*/


    vec<Lit> assumptions;
    bool sat = solver.solve(assumptions);

    // Print out solution
    if (sat) {
        /*std::cout << "Solution: \n";
        for (uint i=0;i<pos.getSizeSigma()*targetNofStates*targetNofStates+targetNofStates;i++) {
            std::cout << " " << i << ": ";
            if (solver.model[i]==l_True) std::cout << "1\n";
            else if (solver.model[i]==l_False) std::cout << "0\n";
            else std::cout << "?\n";
        }*/

        // Build smaller NBW
        NBW newOne(targetNofStates,pos.getSizeSigma());
        for (uint i=0;i<targetNofStates;i++) {
           newOne.setAccepting(i,solver.model[acceptanceVars[i]]==l_True);
           for (uint j=0;j<pos.getSizeSigma();j++) {
                for (uint k=0;k<targetNofStates;k++) {
                    newOne.setTransition(i,j,k,solver.model[transitionVars[i][j][k]]==l_True);
                }
            }
        }
        newOne.copyAlphabetFrom(pos);

        newOne.print();
    } else {
        std::cerr << "No smaller automaton found!\n";
    }

    // Clean up
    for (uint i=0;i<targetNofStates;i++) {
        for (uint j=0;j<pos.getSizeSigma();j++) {
            delete[] transitionVars[i][j];
        }
        delete[] transitionVars[i];
    }
    delete[] transitionVars;

}


int main(int argc, char *argv[])
{
    try {
        // Check if #arguments is correct
        if (argc != 4) {
            usage();
            return 1;
        }

        // Interpreting the arguments given
        uint targetNofStates;
        {
            std::istringstream is(argv[2]);
            is >> targetNofStates;
            if (is.fail()) {
                std::cerr << "Cannot read target number of states. Illegal number given.\n";
                return 1;
            }
        }

        uint bound;
        {
            std::istringstream is(argv[3]);
            is >> bound;
            if (is.fail()) {
                std::cerr << "Cannot read bound. Illegal number given.\n";
                return 1;
            }
        }

        // Reading NBWs
        PreNBW pos((std::string(argv[1])+"/pos.txt").c_str());
        PreNBW neg((std::string(argv[1])+"/neg.txt").c_str());

        // Unify alphabet
        pos.addToAlphabet(neg.getAlphabet());
        neg.addToAlphabet(pos.getAlphabet());

        // Compute actual NBWs
        NBW pos2(pos);
        NBW neg2(neg);

        // Check limits
        if ((pos2.getNofStates()>sizeof(size_t)*8) || (neg2.getNofStates()>32000)) {
            std::cerr << "Error: Input automaton has too many states.\n";
            return 1;
        }

#ifdef TRACK_CAUSE_IN_NOT_TOO_MUCH_CONSTRAINT
        if ((1UL << pos.getAlphabet().size())>(sizeof(NotTooMuchCauseTrackingType)*8)) {
            std::cerr << "Error: Alphabet is too large for this NotTooMuchCauseTrackingType!\n";
            return 1;
        }
#endif

        // std::cout << "NEG:\n";
        // neg2.print();
        // std::cout << "\n";

        // Perform minimization
        minimize(pos2,neg2,bound,targetNofStates);

        // All done!
        return 0;
    } catch (GeneralException e) {
        std::cerr << "Exception: " << e.getMessage() << "\n";
        return 1;
    }
}
